package org.coderu.core.impl.explorer;

import static org.coderu.common.utils.CollectionUtils.subtract;
import static org.coderu.common.utils.CollectionUtils.transform;

import java.io.File;
import java.util.Collection;
import java.util.Comparator;

import javax.annotation.Nullable;

import org.coderu.common.utils.CollectionUtils;
import org.coderu.common.utils.FunctionFactory;
import org.coderu.core.api.CompileUnit;
import org.coderu.core.api.Depth;
import org.coderu.core.api.Packagge;
import org.coderu.core.impl.common.types.ClassDependency;
import org.coderu.core.impl.common.types.Package;
import org.coderu.core.impl.common.types.PackageDependency;
import org.coderu.core.impl.common.types.PackageName;
import org.coderu.core.impl.common.types.PackageWithDependencies;
import org.coderu.core.impl.explorer.helpers.HierarchyBuilder;
import org.coderu.core.impl.explorer.helpers.PackageToClassDependency;
import org.coderu.core.impl.explorer.helpers.Packages2AccessRights;
import org.coderu.core.impl.explorer.parser.PackageLoader;
import org.coderu.core.impl.explorer.rules.AccessRightsFactory;

import com.google.common.base.Predicate;


/**
 * 
 * @author arubinov
 *
 */
public final class DependencyChecker {

    public static Collection<ClassDependency> calculateNotAllowedDependencies(
            final Collection<CompileUnit> classpath, 
            final Collection<Packagge> components, 
            final Collection<Packagge> excluded,
            final Depth depth){

            final Collection<PackageWithDependencies> realPackages = 
                PackageLoader.load( toFile(classpath), toPackage(components), toPackage(excluded));

            final Collection<PackageDependency> existingDependencies = 
                Packages2AccessRights.javaPackages2AccessRights(realPackages);

            final Collection<PackageDependency> allowedDependencies = 
                calculateAllowedDependencies(realPackages);

            final PackageName commonParent = PackageName.getCommonParent(toPackage(components));
            
            final Depth absoluteDepth = Depth.add(new Depth(commonParent.getName().size()),depth);
            
            final Collection<PackageDependency> notAllowedDependencies = 
                CollectionUtils.filter(
                    subtract(existingDependencies, allowedDependencies), depthChecker(absoluteDepth));

            final Collection<ClassDependency> classDependencies = 
                CollectionUtils.concat(
                    transform(notAllowedDependencies, PackageToClassDependency.create(realPackages)));

            return FunctionFactory.sorter(ACCESS_RIGHT_COMPARATOR).apply(classDependencies);
        }

        private static Predicate<PackageDependency> depthChecker(final Depth depth){
            
            return new Predicate<PackageDependency>() {
                @Override
                public boolean apply(@Nullable PackageDependency o) {
                
                    if(depth == Depth.UNLIMITED){
                        return true;
                    }else{
                        
                        final PackageName commonParent = 
                            PackageName.getCommonParent(o.getFirst(), o.getSecond());
                        
                        return commonParent.getName().size() < depth.getValue();
                        
                    }
                }
            };
        }

    private static Collection<File> toFile(Collection<CompileUnit> classpath) {
        return transform(classpath, CompileUnit.TO_FILE);
    }

    private static Collection<PackageName> toPackage(final Collection<Packagge> components) {
        return transform( components, PackageName.FROM_API);
    }

    private static Collection<PackageDependency> calculateAllowedDependencies(
        final Collection<PackageWithDependencies> jDependPackages) {

        final Collection<PackageName> packages = 
            Packages2AccessRights.javaPackages2packageIds(jDependPackages);

        final Collection<Package> packageHierarchy = HierarchyBuilder.build(packages);

        return AccessRightsFactory.generateAccessRights(packageHierarchy);
    }


    private static final Comparator<ClassDependency> ACCESS_RIGHT_COMPARATOR = 
        new Comparator<ClassDependency>() {

        public int compare(ClassDependency o1, ClassDependency o2) {
            final int compareFirst = o1.getFirst().compareTo(o2.getFirst());
            if(compareFirst != 0 ){
                return compareFirst;
            }
            final int compareSecond = o1.getSecond().compareTo(o2.getSecond());
            return compareSecond;
        }
    };

}
