package org.coderu.core.impl.explorer.helpers;

import static org.coderu.common.utils.CollectionUtils.*;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.coderu.common.utils.CollectionUtils;
import org.coderu.common.utils.FunctionFactory;
import org.coderu.core.impl.common.types.Package;
import org.coderu.core.impl.common.types.PackageName;

import com.google.common.base.Function;
import static com.google.common.base.Predicates.*;
import com.google.common.collect.Collections2;


public final class HierarchyBuilder {

    public static Collection<Package> build(final Collection<PackageName> nodes){

        final List<Package> allPackages = new LinkedList<Package>();

        // parent first order
        final List<PackageName> sorted = FunctionFactory.sorter(DESCEND_COMPARATOR).apply(nodes);

        for(PackageName packageName: concat(Collections.singletonList(PackageName.ROOT), sorted)){
            final Package packagge = new Package(packageName);
            createParentRelation(packagge,allPackages);
            // child first order
            allPackages.add(0, packagge);
        }
        return transform( 
                Collections2.filter(allPackages, Package.IS_ROOT), 
                removeEmptyPrefixPackages( new HashSet<PackageName>(nodes) ) );
    }

    private static final Function<Package,Package> removeEmptyPrefixPackages(
            final Set<PackageName> allPackages){

        return new Function<Package, Package>() {
            public Package apply(Package arg0) {

                return removeEmptyPrefixPackages(arg0, allPackages);
            }
        };
    }

    private static Package removeEmptyPrefixPackages(Package packagee, Set<PackageName> allPackages){

        if(packagee.getChilds().size() == 1 && isPseudoPackage(packagee, allPackages)){
            final Package child = CollectionUtils.col2one(packagee.getChilds());
            child.setParent(null);
            return removeEmptyPrefixPackages( child, allPackages);
        }
        else{
            return packagee;
        }
    }

    private static void createParentRelation(Package packagge, List<Package> packages) {

        final Package ancestor = findAncestor(packagge,packages);
        if(ancestor != null){
            final Collection<Package> createdPackages = buildHierarchyChain(ancestor,packagge);
            // child first order
            packages.addAll(0, createdPackages);
        }
    }

    /**
     * @return created items in child first order
     */
    private static Collection<Package> buildHierarchyChain(Package ancestor, Package child) {
        if(ancestor.isParentOf(child)){
            setParentRelation(ancestor, child);
            return Collections.emptyList();
        }
        else{
            final PackageName parentId = PackageName.calculateParent(child.getId());
            final Package parent = new Package(parentId);
            setParentRelation(parent, child);
            return concat( Collections.singletonList(parent), buildHierarchyChain(ancestor, parent));
        }
    }

    private static void setParentRelation(final Package parent, final Package child) {
        parent.getChilds().add(child);
        child.setParent(parent);
    }

    private static Package findAncestor(Package packagge, List<Package> packages) {

        return findFirst(packages, compose(PackageName.isAncestorOf(packagge.getId()), Package.TO_ID));
    }

    private static final Comparator<PackageName> DESCEND_COMPARATOR = new Comparator<PackageName>() {

        public int compare(PackageName o1, PackageName o2) {
            return o1.getFullName().compareTo(o2.getFullName());
        }
    };

    private static boolean isPseudoPackage(final Package packagee, Set<PackageName> allPackages) {
        return !allPackages.contains(packagee.getId());
    }

    private HierarchyBuilder() {}
}
